"""
Utilities for exporting data
"""

##
# Imports

from typer import Typer

#

import warnings

import os
from glob import glob
from pathlib import (
    Path,
)
import gzip

import json, yaml
from dataclasses import dataclass

from tqdm import tqdm
import numpy as np
import webdataset as wds

#

import toile.schema as schema
from ._common import (
    _Pathable,
)
from .tiff_import import (
    load_tiff,
    _FilenameParser,
    _make_filename_parser,
)

#

from typing import (
    TypeAlias,
    Literal,
    Optional,
    Sequence,
)


##
# Type shortcuts

_WDSWriter: TypeAlias = wds.writer.ShardWriter | wds.writer.TarWriter


##
# Helper methods

def _write_movie_frames(
            ds: schema.Movie,
            dest: _WDSWriter,
            key_template: Optional[str] = None,
            i_start: int = 0,
        ) -> int:
    """
    TODO
    """
    ##

    # Normalize args
    if key_template is None:
        key_template = 'sample{i:06d}'

    #

    movie_metadata = (
        dict() if ds.metadata is None
        else { k: v
               for k, v in ds.metadata.items() }
    )
    frame_metadata = (
        ds.frame_metadata if ds.frame_metadata is not None
        else [ None for _ in range( ds.frames.shape[0] ) ]
    )
    
    i_dataset = i_start
    for i_movie, cur_frame_meta in (
        zip(
            range( ds.frames.shape[0] ),
            frame_metadata,
        )
    ):
        cur_metadata = { k: v for k, v in movie_metadata.items() }
        cur_metadata['frame'] = cur_frame_meta

        cur_sample = schema.Frame(
            image = ds.frames[i_movie, :, :],
            metadata = cur_metadata,
        )
        dest_data = cur_sample.as_wds
        dest_data['__key__'] = key_template.format(
            i_dataset = i_dataset,
            i_group = i_movie,
        )

        dest.write( dest_data )
        i_dataset += 1
    
    return i_dataset


## Config parsing

@dataclass
class ExportConfig:
    """TODO"""
    ##

    # Required
    inputs: list[str]
    """TODO"""

    # Optional
    output_stem: str | None = None
    """TODO"""
    shard_size: int = 850_000_000
    """TODO"""
    to_uint8: bool = False
    """TODO"""
    compressed: bool = False
    """TODO"""

    filename_parser: _FilenameParser | None = None
    """TODO"""

def _parse_config( input_path: _Pathable ) -> ExportConfig:

    with open( input_path, 'r' ) as f:
        ret_data = yaml.safe_load( f )

    if 'filename_spec' in ret_data:
        # Cache
        filename_spec = ret_data['filename_spec']
        del ret_data['filename_spec']
    else:
        filename_spec = None
    
    ret = ExportConfig( **ret_data )

    if filename_spec is not None:
        ret.filename_parser = _make_filename_parser(
            filename_spec['template'],
            filename_spec['transforms'],
        )

    return ret


##
# Common

ExportKind: TypeAlias = Literal[
    'movies',
    'frames',
]

def export_tiffs(
        _inputs: Sequence[_Pathable],
        _output_dir: _Pathable,
        _stem: str | None = None,
        #
        kind: ExportKind = 'movies',
        to_uint8: bool = False,
        filename_parser: _FilenameParser | None = None,
        #
        shard_size: float = 38_000_000.,
        compressed: bool = False,
        #
        verbose: bool = False,
        #
        **kwargs
    ) -> None:
    """TODO"""
    ##

    def _printv( *a, **b ):
        if verbose:
            print( *a, **b )

    # Normalize args
    inputs: list[Path] = [ Path( p )
                           for p in _inputs ]
    output_dir = Path( _output_dir )
    stem = (
        Path( output_dir ).stem if _stem is None
        else _stem
    )

    if kind == 'frames':
        key_template = 'tseries-{i_dataset}-frame-{i_group}'
    else:
        # TODO Make explicit for other types
        key_template = 'sample-{i_dataset}-{i_group}'

    # Setup output directory
    output_dir.mkdir( parents = True, exist_ok = True )
    output_pattern = (output_dir / f'{stem}-%06d.tar').as_posix()

    # Parse input globs
    input_globs = [ glob( p.as_posix() )
                    for p in inputs ]
    
    input_paths = []
    for g in input_globs:
        input_paths += [ Path( p )
                         for p in g ]

    # Start building dataset
    n_succeeded = 0
    n_failed = 0

    with wds.writer.ShardWriter( output_pattern,
        maxsize = shard_size,
    ) as dest:
        
        for i_input, cur_input_path in enumerate( input_paths ):
            cur_input_path = Path( cur_input_path )

            _printv( f'🤔 Working on {cur_input_path} ...' )

            #
            _printv( '    💽 Loading ...', end = '' )

            try:
                cur_ds = load_tiff( cur_input_path,
                    to_uint8 = to_uint8,
                    filename_parser = filename_parser,
                )
                _printv( ' Done 🟢' )
            
            except Exception as e:
                _printv( ' Failed 🔴' )
                _printv( 8 * ' ', e )
                if not verbose:
                    print( f'Failed to load movie {cur_input_path}:' )
                    print( 4 * ' ', e )

                n_failed += 1
                continue

            #
            _printv( '    📝 Writing to archive ...', end = '' )

            try:

                if kind == 'movies':
                    raise NotImplementedError()
                    # _write_movie_entire( cur_ds, i_sample, dest )
                    # cur_final = i_sample + 1

                elif kind == 'frames':
                    cur_final = _write_movie_frames( cur_ds, dest,
                        key_template = key_template,
                        # i_start = i_dataset,
                    )

                elif kind == 'clips':
                    raise NotImplementedError()
                
                else:
                    raise ValueError( f'Unrecognized export kind: {kind}' )
                
                _printv( ' Done 🟢' )
            
            except Exception as e:
                _printv( ' Failed 🔴' )
                _printv( 8 * ' ', e )
                if not verbose:
                    print( f'Failed to export movie {cur_input_path}:' )
                    print( 4 * ' ', e )

                n_failed += 1
                continue

            #
        
            _printv( '    ✅ Done.' )
            n_succeeded += 1

##

def export_test(
        output_dir: _Pathable,
        stem: str = '',
        #
        kind: ExportKind = 'frames',
        #
        compressed: bool = False,
        #
        **kwargs
    ) -> None:
    """TODO"""

    image_size = (256, 256)
    image_planes = 900

    if len( stem ) == 0:
        stem = Path( output_dir ).stem
    
    Path( output_dir ).mkdir( parents = True, exist_ok = True )

    # wds_extension = '.tar.gz' if compressed else '.tar'
    wds_extension = '.tar'

    wds_pattern = (
        Path( output_dir )
        / f'{stem}-%06d{wds_extension}'
    ).as_posix()

    if kind == 'frames':

        print( 'Exporting frames ...' )

        with wds.writer.ShardWriter( wds_pattern, **kwargs ) as sink:
            for i in tqdm( range( image_planes ) ):
                cur_frame = schema.ImageSample(
                    data = np.random.randint( 32,767, size = image_size )
                )
                sink.write( cur_frame.as_wds )
    
    else:
        raise NotImplementedError()

    if compressed:

        print( 'Compressing outputs ...' )

        shard_glob = (
            Path( output_dir )
            / f'{stem}-*{wds_extension}'
        ).as_posix()

        for p in glob( shard_glob ):
            print( f'    {p}', end = '', flush = True )
            with open( p, 'rb' ) as f_src:
                p_dest = p + '.gz'
                print( f' -> {p_dest} ...', end = '', flush = True )
                with gzip.open( p_dest, 'wb', compresslevel = 4 ) as f_dest:
                    f_dest.write( f_src.read() )
                print( ' Done.' )
            
            os.remove( p )
    
    print( 'Done' )

##
# Typer app

app = Typer()

@app.command( 'test-frames' )
def _cli_export_test_frames(
            output: str,
            stem: str = '',
            compressed: bool = False,
        ):
    export_test( output, stem,
        compressed = compressed,
        #
        kind = 'frames',
    )

def _standardize_config_args(
                input: _Pathable,
                stem: str = '',
                #
                shard_size: int = -1,
                pds: bool = False,
                #
                uint8: bool = False,
                compressed: bool = False,
            ) -> ExportConfig:
    
    if shard_size < 0:
        # Set default shard sizes based on target

        if pds:
            # Limit for Bsky default PDS blob limit
            shard_size = 38_000_000
        else:
            # Keep to `wds` standard of ≤ 1GB per shard (with overhead)
            shard_size = 850_000_000

    input_path = Path( input )
    if input_path.suffix in ('.yaml', '.yml'):
        ret = _parse_config( input_path )

    else:
        ret = ExportConfig(
            inputs = [ input_path.as_posix() ],
            output_stem = None if len( stem ) == 0 else stem,
            shard_size = shard_size,
            to_uint8 = uint8,
            compressed = compressed,
        )
    
    return ret

# @app.command( 'movies' )
# def _cli_export_movies(
#             input: str,
#             output: str,
#             stem: str = '',
#         ):
#     export_tiffs( input, output, stem, kind = 'movies' )

@app.command( 'frames' )
def _cli_export_frames(
            input: Path,
            output: Path,
            stem: str = '',
            #
            shard_size: int = -1,
            pds: bool = False,
            #
            uint8: bool = False,
            compressed: bool = False,
            #
            verbose: bool = False,
        ):
    
    config = _standardize_config_args(
        input, stem, shard_size, pds, uint8, compressed,
    )

    # TODO Implement compresison
    if config.compressed:
        warnings.warn( '* Compression not yet implemented' )

    export_tiffs(
        config.inputs,
        output,
        config.output_stem,
        #
        to_uint8 = config.to_uint8,
        shard_size = float( config.shard_size ),
        filename_parser = config.filename_parser,
        #
        verbose = verbose,
        #
        kind = 'frames'
    )


##


"""TODO"""

##
# Imports

import os
from pathlib import Path
from glob import glob
import warnings
from datetime import datetime
from uuid import UUID
import re

import numpy as np
import skimage.io as skio
import tifffile
import xmltodict

from toile.schema import (
    Movie,
)

from typing import (
    Optional,
    Callable,
    Any,
    TypeAlias,
)

_FilenameParser: TypeAlias = Callable[[str], dict[str, Any]]

from ._common import (
    _Pathable,
    suppress_stderr,
)


##
# Helpers

# Routines for collating OME TIFF metadata

def _collate_frame_metadata( raw: dict[str, Any] ) -> dict[str, Any]:
    """
    TODO
    """

    ret = dict()

    #

    for k, k_new in [
        ('@PositionX', 'position_x'),
        ('@PositionY', 'position_y'),
        ('@PositionZ', 'position_z'),
    ]:
        if k in raw:
            ret[k_new] = float( raw[k] )
    
    if '@TheT' in raw:
        ret['t_index'] = int( raw['@TheT'] )
    
    if '@DeltaT' in raw:
        ret['t'] = float( raw['@DeltaT'] )

    if 'UUID' in raw:
        if '#text' in raw['UUID']:
            try:
                ret['uuid'] = str( UUID( raw['UUID']['#text'] ) )
            except Exception as e:
                print( f'Invalid frame UUID: "{raw['UUID']['#text']}"' )

    #

    return ret

def _collate_metadata( raw: dict[str, Any],
        date_format: str = '%Y-%m-%dT%H:%M:%S'
    ) -> dict[str, Any]:
    """
    TODO
    """

    ret = dict()

    #

    ome = raw['OME']

    if '@UUID' in ome:
        try:
            ret['uuid'] = str( UUID( ome['@UUID'] ) )
        except Exception as e:
            print( f'Invalid frame UUID: "{ome["@UUID"]}"' )

    if 'Image' in ome:
        image = ome['Image']

        if 'AcquisitionDate' in image:
            # TODO Better way of validating
            try:
                ret['date_acquired'] = (
                    datetime
                        .strptime( image['AcquisitionDate'], date_format )
                        .strftime( date_format )
                )
            except Exception as e:
                print( f'** Invalid acquisition date: "{image["AcquisitionDate"]}"' )
        
        if 'Pixels' in image:
            pixels = image['Pixels']

            for k, k_new in [
                ('@PhysicalSizeX', 'scale_x'),
                ('@PhysicalSizeY', 'scale_y'),
                ('@PhysicalSizeZ', 'scale_z'),
            ]:
                if k in pixels:
                    ret[k_new] = float( pixels[k] )
                    # TODO Determine
                    ret['scale_unit'] = 'um'
            
            # TODO Determine
            ret['t_unit'] = 's'

            for k, k_new in [
                ('@SizeX', 'size_x'),
                ('@SizeY', 'size_y'),
                ('@SizeZ', 'size_z'),
                ('@SizeT', 'size_t'),
            ]:
                if k in pixels:
                    ret[k_new] = int( pixels[k] )
            
            #

            if 'Channel' in pixels:
                channel = pixels['Channel']

                # TODO Check metadata structure for multi-channel recordings
                if 'channels' not in ret:
                    ret['channels'] = []
                
                new_channel = dict()
                if '@Name' in channel:
                    new_channel['name'] = channel['@Name']
                
                ret['channels'].append( new_channel )
            
            #

            if 'TiffData' in pixels and 'Plane' in pixels:
                frame_iterator = map(
                    lambda x: dict( **x[0], **x[1] ),
                    zip( pixels['TiffData'], pixels['Plane'] )
                )
            elif 'TiffData' in pixels:
                frame_iterator = map(
                    lambda x: dict( **x ),
                    pixels['TiffData']
                )
            elif 'Plane' in pixels:
                frame_iterator = map(
                    lambda x: dict( **x ),
                    pixels['Plane']
                )
            else:
                frame_iterator = None

            if frame_iterator is not None:
                ret['frames'] = []
                for frame_data in frame_iterator:
                    ret['frames'].append( _collate_frame_metadata( frame_data ) )

    #

    return ret

def _unformat( string, pattern ):
    """TODO"""
    regex = re.sub( r'{(.+?)}', r'(?P<_\1>.+)', pattern )
    search_res = re.search( regex, string )
    assert search_res is not None
    values = list( search_res.groups() )
    keys = re.findall(r'{(.+?)}', pattern)
    _dict = dict(zip(keys, values))
    return _dict

_filename_parser_transforms = {
    'identity': lambda d, k, x: dict( **d, **{ k: x } ),
    'float': lambda d, k, x: dict( **d, **{ k: float( x ) } ),
    'int': lambda d, k, x: dict( **d, **{ k: int( x ) } ),
    'split_age_sex': lambda d, k, x: (
        dict( **d,
            **{
                k.split( '_' )[0]: int( x[:-1] ),
                k.split( '_' )[1]: x[-1],
            }
        )
    ),
    'date_compact': lambda d, k, x: dict( **d, **{ k: datetime.strptime( x, '%Y%m%d' ).isoformat() } ),
}

def _make_filename_parser( template: str, transforms: dict[str, str] ) -> _FilenameParser:
    """TODO"""
    ##
    def _ret( x: str ) -> dict[str, Any]:
        vals_raw = _unformat( x, template )

        vals = dict()
        vals['_source_filename'] = x
        for k, cur_transform in transforms.items():
            if k not in vals_raw:
                continue
            cur_v = vals_raw[k]
            vals = _filename_parser_transforms[cur_transform](vals, k, cur_v )
        
        return vals
    
    return _ret


##
# Main routine

def load_tiff( path: _Pathable,
        frame_pattern_full: str = '*_*0001.ome.tif*',
        frame_pattern: str = '*.ome.tif*',
        filename_parser: Optional[_FilenameParser] = None,
        #
        to_uint8: bool = False,
    ) -> Movie:

    # Normalize args
    path = Path( path )

    if filename_parser is None:
    #     # TODO More general default behavior
        filename_parser = lambda x: dict( filename = x )
        # filename_parser = _parse_filename_1

    #

    # Try full-stack load
    raw_input_full = glob( frame_pattern_full, root_dir = path )
    first_frame_path = None

    # try:

    #
    if len( raw_input_full ) == 1:
        first_frame_path = path / raw_input_full[0]
        
    elif len( raw_input_full ) > 1:

        if all( 'Ch' in x.split( '_' )[-2]
                for x in raw_input_full ):
            first_frame_path = path / raw_input_full[0]

        else:
            raise RuntimeError( f'Unsupported multi-channel format in {path.as_posix()}' )
    
    else:
        raise RuntimeError( f'No matching image stack for {(path / frame_pattern_full).as_posix()}' )

    #

    assert first_frame_path is not None

    #

    # We suppress stderr to hopefully avoid scikit-image's nonsense
    with suppress_stderr():

        first_frame_filename = first_frame_path.name
        # print( first_frame_filename )
        # print( os.path.split( first_frame_filename ) )
        filename_metadata = filename_parser( os.path.split( first_frame_filename )[1] )
        # from pprint import pprint
        # pprint( filename_metadata )
        stack = skio.imread( first_frame_path )
    
    # except Exception as e:
    #     print( f'Issue with:' )
    #     print( raw_input_full )
    #     # Re-raise
    #     raise e

    #

    if to_uint8:
        # Perform stack-wide image normalization
        max_value = np.max( stack )
        tmp = (255. / max_value) * stack.astype( float )
        tmp = np.floor( tmp )
        stack = tmp.astype( np.uint8 )

    #

    # filename_metadata = filename_parser( os.path.split( path )[1] )

    with warnings.catch_warnings():
        warnings.simplefilter( 'ignore' )

        with tifffile.TiffFile( first_frame_path ) as tif:
            if tif.ome_metadata is not None:
                image_metadata = _collate_metadata( xmltodict.parse( tif.ome_metadata ) )
            else:
                image_metadata = dict()
    
    frame_metadata = image_metadata.get( 'frames', None )
    del image_metadata['frames']

    movie_metadata = dict(
        **filename_metadata,
        **image_metadata,
    )

    #

    return Movie(
        frames = stack,
        metadata = movie_metadata,
        frame_metadata = frame_metadata,
    )


#


"""
TODO
"""

from typer import Typer

from .export import app as export_app


##

app = Typer()

app.add_typer( export_app, name = 'export' )


##

def main():
    app()


##

if __name__ == '__main__':
    main()


##


"""
Dataset sample schemas
"""

##
# Imports

import atdata

from dataclasses import dataclass

from typing import (
    Any,
    Literal,
    TypeAlias,
)
from numpy.typing import (
    NDArray,
)


##
# Sample types

# Helpers

PositionUnit: TypeAlias = Literal[
    'um',
    'mm',
    'm',
]
TimeUnit: TypeAlias = Literal[
    'ms',
    's',
    'min',
    'h'
]

Identifier: TypeAlias = int | str

## Convenience stores for intermediate handling

@dataclass
class Movie( atdata.PackableSample ):
    """Generic movie data extracted from a TIFF stack"""
    frames: NDArray
    metadata: dict[str, Any] | None = None
    frame_metadata: list[dict[str, Any]] | None = None

@dataclass
class Frame( atdata.PackableSample ):
    """Generic image data extracted as a movie frame from a TIFF stack"""
    image: NDArray
    metadata: dict[str, Any] | None = None

## NEW

@dataclass
class SliceRecordingFrame( atdata.PackableSample ):
    """TODO"""

    ##

    # Required
    data: NDArray
    """`numpy` array containing the frame's image data"""
    
    mouse_id: Identifier
    """TODO"""
    slice_id: Identifier
    """TODO"""

@dataclass
class ImageSample( atdata.PackableSample ):
    """TODO"""
    data: NDArray

@atdata.lens
def project_image( source: SliceRecordingFrame ) -> ImageSample:
    return ImageSample(
        data = source.data
    )
@project_image.putter
def put_image( view: ImageSample, source: SliceRecordingFrame ) -> SliceRecordingFrame:
    return SliceRecordingFrame(
        data = view.data,
        mouse_id = source.mouse_id,
        slice_id = source.slice_id
    )

## OLD

@dataclass
class Position3:
    """TODO"""
    ##

    # Required
    x: float
    """The $x$-position"""
    y: float
    """The $y$-position"""
    z: float
    """The $z$-position"""

    # Optional
    unit: PositionUnit | None = None
    """Unit for interpreting coordinate vlaues"""

# Metadata

@dataclass
class MovieFrameMetadata:
    """TODO"""
    ##

    # Required
    t_index: int
    """Sequential index of this frame within the larger recording"""

    # Optional
    position: Position3 | None = None
    """The offset position of the stage at this time in the recording"""
    t: float | None = None
    """Acquisition time of this frame (in s)"""
    uuid: str | None = None
    """UUID given to frame at acquisition"""

@dataclass
class MovieMetadata:
    """TODO"""

    ##

    # Required
    # ...

    # Optional
    filename: str | None = None
    """Original source filename of raw movie"""
    date_saved: str | None = None
    """Timestamp for when the original full movie file was saved"""

@dataclass
class ChannelMetadata:
    """TODO"""

    ##

    # Required
    # ...

    # Optional
    name: str | None = None
    """Descriptive name of this recording channel"""

@dataclass
class SliceRecordingMetadata:
    """TODO"""

    ##

    # Required
    mouse_id: Identifier
    """TODO"""
    slice_id: Identifier
    """TODO"""

    # Optional
    intervention: Identifier | None = None
    """TODO"""
    condition: Identifier | None = None
    """TODO"""
    replicate_id: Identifier | None = None
    """TODO"""


# Samples

# @atdata.packable
# class SliceRecordingFrame:
#     """TODO"""

#     ##

#     # Required
#     data: NDArray
#     """`numpy` array containing the frame's image data"""

#     session: SliceRecordingMetadata
#     """Metadata about the experimental session"""
#     movie: MovieMetadata
#     """Metadata about the full movie recording"""
#     frame: MovieFrameMetadata
#     """Metadata about this individual frame within the full recording"""

# @dataclass
# class SliceRecordingFrame( atdata.PackableSample ):
#     """TODO"""

#     ##

#     # Required
#     data: NDArray
#     """`numpy` array containing the frame's image data"""
    
#     mouse_id: Identifier
#     """TODO"""
#     slice_id: Identifier
#     """TODO"""


#


"""TODO"""

##
# Imports

import os, sys
from pathlib import Path


##
# Typing shortcuts

_Pathable = str | Path


##
# Contexts

class _SuppressStderrContext:
    """Suppress error printing within the context
    
    USE WITH CAUTION.
    """

    def __init__( self ):
        """Construct a new context manager"""
        pass

    def __enter__( self ):
        sys.stderr = open( os.devnull, 'w' )

    def __exit__( self, exc_type, exc_val, exc_tb ):
        sys.stderr = sys.__stderr__

def suppress_stderr() -> _SuppressStderrContext:
    return _SuppressStderrContext()


#